import inspect

import numpy as np
from scipy.optimize import minimize

import skgstat as skg
from skgstat import models
import skgstat.util.likelihood as li


def test_wrapped_model_doc():
    """Test the docstring wrapping"""
    # create a wrapped spherical function
    wrapped = li._model_transformed(models.spherical, has_s=False)

    assert 'Autocorrelation function.' in wrapped.__doc__
    assert models.spherical.__doc__ in wrapped.__doc__


def test_wrapped_model_args():
    """"
    Check that the number of model parameters is initialized correctly.
    """
    # get the two functions
    spherical = li._model_transformed(models.spherical, has_s=False)
    stable = li._model_transformed(models.stable, has_s=True)

    sig = inspect.signature(spherical)
    assert len(sig.parameters) == 2

    sig = inspect.signature(stable)
    assert len(sig.parameters) == 3


def test_build_A():
    """
    Test the autocorrelation matrix building.
    """
    # create a wrapped spherical function
    wrap_2 = li._model_transformed(models.spherical, has_s=False)
    wrap_3 = li._model_transformed(models.stable, has_s=True)

    # build the autocorrelation matrix
    A_5 = li._build_A(wrap_2, [1, 1, 0], np.arange(0, 1, 0.1))
    A_6 = li._build_A(wrap_3, [1, 1, 1, 1], np.arange(0, 1.5, 0.1))

    # check the matrix shape
    assert A_5.shape == (5, 5)
    assert A_6.shape == (6, 6)


def test_likelihood():
    """
    Call the likelihood function and make sure that it optimizes the
    the pancake variogram
    """
    # build the variogram from the tutorial
    c, v = skg.data.pancake(300, seed=42).get('sample')
    vario = skg.Variogram(c, v, bin_func='scott', maxlag=0.7)

    # get the likelihood function
    like = li.get_likelihood(vario)

    # cretae the optimization attributes
    sep_mean = vario.distance.mean()
    sam_var = vario.values.var()

    # create initial guess
    p0 = np.array([sep_mean, sam_var, 0.1 * sam_var])

    # create the bounds to restrict optimization
    bounds = [[0, vario.bins[-1]], [0, 3*sam_var], [0, 2.9*sam_var]]

    # minimize the likelihood function
    res = minimize(like, p0, bounds=bounds, method='SLSQP')

    # the result and p0 should be different
    assert not np.allclose(res.x, p0, rtol=1e-3)



from typing import Tuple, List

from scipy.spatial.distance import pdist, cdist, squareform
from scipy.spatial import cKDTree
from scipy import sparse
import numpy as np
import multiprocessing as mp


def _sparse_dok_get(m, fill_value=np.NaN):
    """Like m.toarray(), but setting empty values to `fill_value`, by
    default `np.NaN`, rather than 0.0.

    Parameters
    ----------
    m : scipy.sparse.dok_matrix
    fill_value : float
    """
    mm = np.full(m.shape, fill_value)
    for (x, y), value in m.items():
        mm[x, y] = value
    return mm


class DistanceMethods(object):
    def find_closest(self, idx, max_dist=None, N=None):
        """find neighbors
        Find the (N) closest points (in the right set) to the point with
        index idx (in the left set).

        Parameters
        ----------
        idx : int
            Index of the point that the N closest neighbors
            are searched for.
        max_dist : float
            Maximum distance at which other points are searched
        N : int
            Number of points searched.

        Returns
        -------
        ridx : numpy.ndarray
            Indices of the N closeset points to idx

        """

        if max_dist is None:
            max_dist = self.max_dist
        else:
            if self.max_dist is not None and max_dist != self.max_dist:
                raise AttributeError(
                    "max_dist specified and max_dist != self.max_dist"
                )

        if isinstance(self.dists, sparse.spmatrix):
            dists = self.dists.getrow(idx)
        else:
            dists = self.dists[idx, :]
        if isinstance(dists, sparse.spmatrix):
            ridx = np.array([k[1] for k in dists.todok().keys()])
        elif max_dist is not None:
            ridx = np.where(dists <= max_dist)[0]
        else:
            ridx = np.arange(len(dists))
        if ridx.size > N:
            if isinstance(dists, sparse.spmatrix):
                selected_dists = dists[0, ridx].toarray()[0, :]
            else:
                selected_dists = dists[ridx]
            sorted_ridx = np.argsort(selected_dists, kind="stable")
            ridx = ridx[sorted_ridx][:N]
        return ridx


class MetricSpace(DistanceMethods):
    """
    A MetricSpace represents a point cloud together with a distance
    metric and possibly a maximum distance. It efficiently provides
    the distances between each point pair (when shorter than the
    maximum distance).

    Note: If a max_dist is specified a sparse matrix representation is
    used for the distances, which saves space and calculation time for
    large datasets, especially where max_dist << the size of the point
    cloud in space. However, it slows things down for small datasets.
    """

    def __init__(self, coords, dist_metric="euclidean", max_dist=None):
        """ProbabalisticMetricSpace class

        Parameters
        ----------
        coords : numpy.ndarray
            Coordinate array of shape (Npoints, Ndim)
        dist_metric : str
            Distance metric names as used by scipy.spatial.distance.pdist
        max_dist : float
            Maximum distance between points after which the distance
            is considered infinite and not calculated.
        """
        self.coords = coords.copy()
        self.dist_metric = dist_metric
        self.max_dist = max_dist
        self._tree = None
        self._dists = None

        # Check if self.dist_metric is valid
        try:
            pdist(self.coords[:1, :], metric=self.dist_metric)
        except ValueError as e:
            raise e

    @property
    def tree(self):
        """If `self.dist_metric` is `euclidean`, a `scipy.spatial.cKDTree`
        instance of `self.coords`. Undefined otherwise."""
        # only Euclidean supported
        if self.dist_metric != "euclidean":
            raise ValueError((
                "A coordinate tree can only be constructed "
                "for an euclidean space"
            ))

        # if not cached - calculate
        if self._tree is None:
            self._tree = cKDTree(self.coords)

        # return
        return self._tree

    @property
    def dists(self):
        """A distance matrix of all point pairs. If `self.max_dist` is
        not `None` and `self.dist_metric` is set to `euclidean`, a
        `scipy.sparse.csr_matrix` sparse matrix is returned.
        """
        # calculate if not cached
        if self._dists is None:
            # check if max dist is given
            if self.max_dist is not None and self.dist_metric == "euclidean":
                self._dists = self.tree.sparse_distance_matrix(
                    self.tree,
                    self.max_dist,
                    output_type="coo_matrix"
                ).tocsr()

            # otherwise use pdist
            else:
                self._dists = squareform(
                    pdist(self.coords, metric=self.dist_metric)
                )

        # return
        return self._dists

    def diagonal(self, idx=None):
        """
        Return a diagonal matrix (as per
        :func:`squareform <scipy.spatial.distance.squareform>`),
        optionally for a subset of the points

        Parameters
        ----------
        idx : list
            list of indices that the diagonal matrix is calculated for.

        Returns
        -------
        diagonal : numpy.ndarray
            squareform matrix of the subset of coordinates

        """
        # get the dists
        dist_mat = self.dists

        # subset dists if requested
        if idx is not None:
            dist_mat = dist_mat[idx, :][:, idx]

        # handle sparse matrix
        if isinstance(self.dists, sparse.spmatrix):
            dist_mat = _sparse_dok_get(dist_mat.todok(), np.inf)
            np.fill_diagonal(dist_mat, 0)  # Normally set to inf

        return squareform(dist_mat)

    def __len__(self):
        return len(self.coords)


class MetricSpacePair(DistanceMethods):
    """
    A MetricSpacePair represents a set of point clouds (MetricSpaces).
    It efficiently provides the distances between each point in one
    point cloud and each point in the other point cloud (when shorter
    than the maximum distance). The two point clouds are required to
    have the same distance metric as well as maximum distance.
    """
    def __init__(self, ms1, ms2):
        """
        Parameters
        ----------
        ms1 : MetricSpace
        ms2 : MetricSpace

        Note: `ms1` and `ms2` need to have the same `max_dist` and
        `distance_metric`.
        """
        # check input data
        # same distance metrix
        if ms1.dist_metric != ms2.dist_metric:
            raise ValueError(
                "Both MetricSpaces need to have the same distance metric"
            )

        # same max_dist setting
        if ms1.max_dist != ms2.max_dist:
            raise ValueError(
                "Both MetricSpaces need to have the same max_dist"
            )
        self.ms1 = ms1
        self.ms2 = ms2
        self._dists = None

    @property
    def dist_metric(self):
        return self.ms1.dist_metric

    @property
    def max_dist(self):
        return self.ms1.max_dist

    @property
    def dists(self):
        """A distance matrix of all point pairs. If `self.max_dist` is
        not `None` and `self.dist_metric` is set to `euclidean`, a
        `scipy.sparse.csr_matrix` sparse matrix is returned.
        """
        # if not cached, calculate
        if self._dists is None:
            # handle euclidean with max_dist with Tree
            if self.max_dist is not None and self.dist_metric == "euclidean":
                self._dists = self.ms1.tree.sparse_distance_matrix(
                    self.ms2.tree,
                    self.max_dist,
                    output_type="coo_matrix"
                ).tocsr()

            # otherwise Tree not possible
            else:
                self._dists = cdist(
                    self.ms1.coords,
                    self.ms2.coords,
                    metric=self.ms1.dist_metric
                )

        # return
        return self._dists


class ProbabalisticMetricSpace(MetricSpace):
    """Like MetricSpace but samples the distance pairs only returning a
       `samples` sized subset. `samples` can either be a fraction of
       the total number of pairs (float < 1), or an integer count.
    """
    def __init__(
            self,
            coords,
            dist_metric="euclidean",
            max_dist=None,
            samples=0.5,
            rnd=None
        ):
        """ProbabalisticMetricSpace class

        Parameters
        ----------
        coords : numpy.ndarray
            Coordinate array of shape (Npoints, Ndim)
        dist_metric : str
            Distance metric names as used by scipy.spatial.distance.pdist
        max_dist : float
            Maximum distance between points after which the distance
            is considered infinite and not calculated.
        samples : float, int
            Number of samples (int) or fraction of coords to sample (float < 1).
        rnd : numpy.random.RandomState, int
            Random state to use for the sampling.
        """
        self.coords = coords.copy()
        self.dist_metric = dist_metric
        self.max_dist = max_dist
        self.samples = samples
        if rnd is None:
            self.rnd = np.random
        elif isinstance(rnd, np.random.RandomState):
            self.rnd = rnd
        else:
            self.rnd = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(rnd)))

        self._lidx = None
        self._ridx = None
        self._ltree = None
        self._rtree = None
        self._dists = None
        # Do a very quick check to see throw exceptions 
        # if self.dist_metric is invalid...
        pdist(self.coords[:1, :], metric=self.dist_metric)

    @property
    def sample_count(self):
        if isinstance(self.samples, int):
            return self.samples
        return int(self.samples * len(self.coords))

    @property
    def lidx(self):
        """The sampled indices into `self.coords` for the left sample."""
        if self._lidx is None:
            self._lidx = self.rnd.choice(len(self.coords), size=self.sample_count, replace=False)
        return self._lidx

    @property
    def ridx(self):
        """The sampled indices into `self.coords` for the right sample."""
        if self._ridx is None:
            self._ridx = self.rnd.choice(len(self.coords), size=self.sample_count, replace=False)
        return self._ridx

    @property
    def ltree(self):
        """If `self.dist_metric` is `euclidean`, a `scipy.spatial.cKDTree`
        instance of the left sample of `self.coords`. Undefined otherwise."""

        # only Euclidean supported
        if self.dist_metric != "euclidean":
            raise ValueError((
                "A coordinate tree can only be constructed "
                "for an euclidean space"
            ))

        if self._ltree is None:
            self._ltree = cKDTree(self.coords[self.lidx, :])
        return self._ltree

    @property
    def rtree(self):
        """If `self.dist_metric` is `euclidean`, a `scipy.spatial.cKDTree`
        instance of the right sample of `self.coords`. Undefined otherwise."""

        # only Euclidean supported
        if self.dist_metric != "euclidean":
            raise ValueError((
                "A coordinate tree can only be constructed "
                "for an euclidean space"
            ))

        if self._rtree is None:
            self._rtree = cKDTree(self.coords[self.ridx, :])
        return self._rtree

    @property
    def dists(self):
        """A distance matrix of the sampled point pairs as a
        `scipy.sparse.csr_matrix` sparse matrix. """
        if self._dists is None:
            max_dist = self.max_dist
            if max_dist is None:
                max_dist = np.finfo(float).max
            dists = self.ltree.sparse_distance_matrix(
                self.rtree,
                max_dist,
                output_type="coo_matrix"
            ).tocsr()
            dists.resize((len(self.coords), len(self.coords)))
            dists.indices = self.ridx[dists.indices]
            dists = dists.tocsc()
            dists.indices = self.lidx[dists.indices]
            dists = dists.tocsr()
            self._dists = dists
        return self._dists


# Subfunctions used in RasterEquidistantMetricSpace 
# (outside class so that they can be pickled by multiprocessing)
def _get_disk_sample(
    coords: np.ndarray,
    center: Tuple[float, float],
    center_radius: float,
    rnd_func: np.random.RandomState,
    sample_count: int
):
    """
    Subfunction for RasterEquidistantMetricSpace.
    Calculates the indexes of a subsample in a disk "center sample".
    Same parameters as in the class.
    """
    # First index: preselect samples in a disk of certain radius
    dist_center = np.sqrt((coords[:, 0] - center[0]) ** 2 + (
            coords[:, 1] - center[1]) ** 2)
    idx1 = dist_center < center_radius

    count = np.count_nonzero(idx1)
    indices1 = np.argwhere(idx1)

    # Second index: randomly select half of the valid pixels, 
    # so that the other half can be used by the equidist
    # sample for low distances
    indices2 = rnd_func.choice(count, size=min(count, sample_count), replace=False)

    if count != 1:
        return indices1[indices2].squeeze()
    else:
        return indices1[indices2][0]


def _get_successive_ring_samples(
    coords: np.ndarray,
    center: Tuple[float, float],
    equidistant_radii: List[float],
    rnd_func: np.random.RandomState, sample_count: int
):
    """
    Subfunction for RasterEquidistantMetricSpace.
    Calculates the indexes of several subsamples within disks,
    "equidistant sample". Same parameters as in the class.
    """
    # First index: preselect samples in a ring of certain inside radius and outside radius
    dist_center = np.sqrt((coords[:, 0] - center[0]) ** 2 + (coords[:, 1] - center[1]) ** 2)

    idx = np.logical_and(
        dist_center[None, :] >= np.array(equidistant_radii[:-1])[:, None],
        dist_center[None, :] < np.array(equidistant_radii[1:])[:, None]
    )

    # Loop over an iterative sampling in rings
    list_idx = []
    for i in range(len(equidistant_radii) - 1):
        idx1 = idx[i, :]

        count = np.count_nonzero(idx1)
        indices1 = np.argwhere(idx1)

        # Second index: randomly select half of the valid pixels, so that the other half can be used by the equidist
        # sample for low distances
        indices2 = rnd_func.choice(count, size=min(count, sample_count), replace=False)
        sub_idx = indices1[indices2]

        if count > 1:
            list_idx.append(sub_idx.squeeze())
        elif count == 1:
            list_idx.append(sub_idx[0])

    return np.concatenate(list_idx)


def _get_idx_dists(
    coords: np.ndarray,
    center: Tuple[float, float],
    center_radius: float,
    equidistant_radii: List[float],
    rnd_func: np.random.RandomState,
    sample_count: int,
    max_dist: float,
    i: int,
    imax: int,
    verbose: bool
):
    """
    Subfunction for RasterEquidistantMetricSpace.
    Calculates the pairwise distances between a list of pairs of "center" and "equidistant" ensembles.
    Same parameters as in the class.
    """

    if verbose:
        print('Working on subsample ' + str(i+1) + ' out of ' + str(imax))

    cidx = _get_disk_sample(
        coords=coords, center=center,
        center_radius=center_radius,
        rnd_func=rnd_func,
        sample_count=sample_count
    )

    eqidx = _get_successive_ring_samples(
        coords=coords,
        center=center,
        equidistant_radii=equidistant_radii,
        rnd_func=rnd_func,
        sample_count=sample_count
    )

    ctree = cKDTree(coords[cidx, :])
    eqtree = cKDTree(coords[eqidx, :])

    dists = ctree.sparse_distance_matrix(
        eqtree,
        max_dist,
        output_type="coo_matrix"
    )

    return dists.data, cidx[dists.row], eqidx[dists.col]


def _mp_wrapper_get_idx_dists(argdict: dict):
    """
    Multiprocessing wrapper for get_idx_dists.
    """
    return _get_idx_dists(**argdict)


class RasterEquidistantMetricSpace(MetricSpace):
    """Like ProbabilisticMetricSpace but only applies to Raster data (2D gridded data) and
    samples iteratively an `equidistant` subset within distances to a 'center' subset.
    Subsets can either be a fraction of the total number of pairs (float < 1), or an integer count.
    The 'center' subset corresponds to a disk centered on a point of the grid for which the location
    randomly varies and can be redrawn and aggregated for several runs. The corresponding 'equidistant'
    subset consists of a concatenation of subsets drawn from rings with radius gradually increasing
    until the maximum extent of the grid is reached.

    To define the subsampling, several parameters are available:
    - The raw number of samples corresponds to the samples that will be drawn in each central disk.
     Along with the ratio of samples drawn (see below), it will automatically define the radius
     of the disk and rings for subsampling.
     Note that the number of samples drawn will be repeatedly drawn for each equidistant rings
     at a given radius, resulting in a several-fold amount of total samples for the equidistant
     subset.
    - The ratio of subsample defines the density of point sampled within each subset. It
     defaults to 20%.
    - The number of runs corresponds to the number of random center points repeated during the
    subsampling. It defaults to a sampling of 1% of the grid with center subsets.

    Alternatively, one can supply:
    - The multiplicative factor to derive increasing rings radii, set as squareroot of 2 by
    default in order to conserve a similar area for each ring and verify the sampling ratio.
    Or directly:
    - The radius of the central disk subset.
    - A list of radii for the equidistant ring subsets.
    When providing those spatial parameters, all other sampling parameters will be ignored
    except for the raw number of samples to draw in each subset.
      """

    def __init__(
            self,
            coords,
            shape,
            extent,
            samples=100,
            ratio_subsample=0.2,
            runs=None,
            n_jobs=1,
            exp_increase_fac=np.sqrt(2),
            center_radius=None,
            equidistant_radii=None,
            max_dist=None,
            dist_metric="euclidean",
            rnd=None,
            verbose=False
    ):
        """RasterEquidistantMetricSpace class

        Parameters
        ----------
        coords : numpy.ndarray
            Coordinate array of shape (Npoints, 2)
        shape : tuple[int, int]
            Shape of raster (X, Y)
        extent : tuple[float, float, float, float]
            Extent of raster (Xmin, Xmax, Ymin, Ymax)
        samples : float, int
            Number of samples (int) or fraction of coords to sample (float < 1).
        ratio_subsample:
            Ratio of samples drawn within each subsample.
        runs : int
            Number of subsamplings based on a random center point
        n_jobs : int
            Number of jobs to use in multiprocessing for the subsamplings.
        exp_increase_fac : float
            Multiplicative factor of increasing radius for ring subsets
        center_radius: float
            Radius of center subset, overrides other sampling parameters.
        equidistant_radii: list
            List of radii of ring subset, overrides other sampling parameters.
        dist_metric : str
            Distance metric names as used by scipy.spatial.distance.pdist
        max_dist : float
            Maximum distance between points after which the distance
            is considered infinite and not calculated.
        verbose : bool
            Whether to print statements in the console

        rnd : numpy.random.RandomState, int
            Random state to use for the sampling.
        """

        if dist_metric != "euclidean":
            raise ValueError((
                "A RasterEquidistantMetricSpace class can only be constructed "
                "for an euclidean space"
            ))

        self.coords = coords.copy()
        self.dist_metric = dist_metric
        self.shape = shape
        self.extent = extent
        self.res = np.mean([(extent[1] - extent[0])/(shape[0]-1),(extent[3] - extent[2])/(shape[1]-1)])

        # if the maximum distance is not specified, find the maximum possible distance from the extent
        if max_dist is None:
            max_dist = np.sqrt((extent[1] - extent[0])**2 + (extent[3] - extent[2])**2)
        self.max_dist = max_dist

        self.samples = samples

        if runs is None:
            # If None is provided, try to sample center samples for about one percent of the area
            runs = int((self.shape[0] * self.shape[1]) / self.samples * 1/100.)
        self.runs = runs

        self.n_jobs = n_jobs

        if rnd is None:
            self.rnd = np.random.default_rng()
        elif isinstance(rnd, np.random.RandomState):
            self.rnd = rnd
        else:
            self.rnd = np.random.RandomState(np.random.MT19937(np.random.SeedSequence(rnd)))

        # Radius of center subsample, based on sample count
        # If None is provided, the disk is defined with the exact size to hold the number of percentage of samples
        # defined by the user
        if center_radius is None:
            center_radius = np.sqrt(1. / ratio_subsample * self.sample_count / np.pi) * self.res
            if verbose:
                print('Radius of center disk sample for sample count of '+str(self.sample_count)+ ' and subsampling ratio'
                      ' of '+str(ratio_subsample)+': '+str(center_radius))
        self._center_radius = center_radius

        # Radii of equidistant ring subsamples
        # If None is provided, the rings are defined with exponentially increasing radii with a factor sqrt(2), which
        # means each ring will have just enough area to sample at least the number of samples desired, and same
        # for each of the following, due to:
        # (sqrt(2)R)**2 - R**2 = R**2
        if equidistant_radii is None:
            equidistant_radii = [0.]
            increasing_rad = self._center_radius
            while increasing_rad < self.max_dist:
                equidistant_radii.append(increasing_rad)
                increasing_rad *= exp_increase_fac
            equidistant_radii.append(self.max_dist)
            if verbose:
                print('Radii of equidistant ring samples for increasing factor of ' + str(exp_increase_fac) + ': ')
                print(equidistant_radii)
        self._equidistant_radii = equidistant_radii

        self.verbose = verbose

        # Index and KDTree of center sample
        self._cidx = None
        self._ctree = None

        # Index and KDTree of equidistant sample
        self._eqidx = None
        self._eqtree = None

        self._centers = None
        self._dists = None
        # Do a very quick check to see throw exceptions
        # if self.dist_metric is invalid...
        pdist(self.coords[:1, :], metric=self.dist_metric)

    @property
    def sample_count(self):
        if isinstance(self.samples, int):
            return self.samples
        return int(self.samples * len(self.coords))

    @property
    def cidx(self):
        """The sampled indices into `self.coords` for the center sample."""
        return self._cidx

    @property
    def ctree(self):
        """If `self.dist_metric` is `euclidean`, a `scipy.spatial.cKDTree`
        instance of the center sample of `self.coords`. Undefined otherwise."""

        # only Euclidean supported
        if self.dist_metric != "euclidean":
            raise ValueError((
                "A coordinate tree can only be constructed "
                "for an euclidean space"
            ))

        if self._ctree is None:
            self._ctree = [cKDTree(self.coords[self.cidx[i], :]) for i in range(len(self.cidx))]
        return self._ctree


    @property
    def eqidx(self):
        """The sampled indices into `self.coords` for the equidistant sample."""
        return self._eqidx

    @property
    def eqtree(self):
        """If `self.dist_metric` is `euclidean`, a `scipy.spatial.cKDTree`
        instance of the equidistant sample of `self.coords`. Undefined otherwise."""

        # only Euclidean supported

        if self._eqtree is None:
            self._eqtree = [cKDTree(self.coords[self.eqidx[i], :]) for i in range(len(self.eqidx))]
        return self._eqtree

    @property
    def dists(self):
        """A distance matrix of the sampled point pairs as a
        `scipy.sparse.csr_matrix` sparse matrix. """

        # Derive distances
        if self._dists is None:

            idx_center = self.rnd.choice(len(self.coords), size=min(self.runs, len(self.coords)), replace=False)

            # Each run has a different center
            centers = self.coords[idx_center]

            # Running on a single core: for loop
            if self.n_jobs == 1:

                list_dists, list_cidx, list_eqidx = ([] for i in range(3))

                for i in range(self.runs):

                    center = centers[i]
                    dists, cidx, eqidx = _get_idx_dists(self.coords, center=center, center_radius=self._center_radius,
                                                       equidistant_radii=self._equidistant_radii, rnd_func=self.rnd,
                                                       sample_count=self.sample_count, max_dist=self.max_dist, i=i,
                                                       imax=self.runs, verbose=self.verbose)
                    list_dists.append(dists)
                    list_cidx.append(cidx)
                    list_eqidx.append(eqidx)

            # Running on several cores: multiprocessing
            else:
                # Arguments to pass: only centers and loop index for verbose are changing
                argsin = [{'center': centers[i], 'coords': self.coords, 'center_radius': self._center_radius,
                           'equidistant_radii': self._equidistant_radii, 'rnd_func': self.rnd,
                           'sample_count': self.sample_count, 'max_dist': self.max_dist, 'i': i, 'imax': self.runs,
                           'verbose': self.verbose} for i in range(self.runs)]

                # Process in parallel
                pool = mp.Pool(self.n_jobs, maxtasksperchild=1)
                outputs = pool.map(_mp_wrapper_get_idx_dists, argsin, chunksize=1)
                pool.close()
                pool.join()

                # Get lists of outputs
                list_dists, list_cidx, list_eqidx = list(zip(*outputs))

            # Define class objects
            self._centers = centers
            self._cidx = list_cidx
            self._eqidx = list_eqidx

            # concatenate the coo matrixes
            d = np.concatenate(list_dists)
            c = np.concatenate(list_cidx)
            eq = np.concatenate(list_eqidx)

            # remove possible duplicates (that would be summed by default)
            # from https://stackoverflow.com/questions/28677162/ignoring-duplicate-entries-in-sparse-matrix

            # Stable solution but a bit slow
            # c, eq, d = zip(*set(zip(c, eq, d)))
            # dists = sparse.csr_matrix((d, (c, eq)), shape=(len(self.coords), len(self.coords)))

            # Solution 5+ times faster than the preceding, but relies on _update() which might change in scipy (which
            # only has an implemented method for summing duplicates, and not ignoring them yet)
            dok = sparse.dok_matrix((len(self.coords), len(self.coords)))
            dok._update(zip(zip(c, eq), d))
            dists = dok.tocsr()

            self._dists = dists

        return self._dists


import unittest
import os
import pickle

import numpy as np
import pandas as pd
from numpy.testing import assert_array_almost_equal
from scipy.spatial.distance import pdist

try:
    import plotly.graph_objects as go
    PLOTLY_FOUND = True
except ImportError:
    print('No plotly installed. Skip plot tests')
    PLOTLY_FOUND = False

from skgstat import Variogram, DirectionalVariogram
from skgstat import OrdinaryKriging
from skgstat import estimators
from skgstat import plotting


class TestSpatiallyCorrelatedData(unittest.TestCase):
    def setUp(self):
        # Generate some random but spatially correlated data
        # with a range of ~20
        
        np.random.seed(42)
        c = np.random.sample((50, 2)) * 60
        np.random.seed(42)
        v = np.random.normal(10, 4, 50)
        
        V = Variogram(c, v).describe()
        V["effective_range"] = 20
        OK = OrdinaryKriging(V, coordinates=c, values=v)

        self.c = np.random.sample((500, 2)) * 60
        self.v = OK.transform(self.c)

        self.c = self.c[~np.isnan(self.v),:]
        self.v = self.v[~np.isnan(self.v)]

    def test_dense_maxlag_inf(self):
        Vdense = Variogram(self.c, self.v)
        Vsparse = Variogram(self.c, self.v, maxlag=10000000)

        for x, y in zip(Vdense.parameters, Vsparse.parameters):
            self.assertAlmostEqual(x, y, places=3)
            
    def test_sparse_maxlag_50(self):
        V = Variogram(self.c, self.v, maxlag=50)

        for x, y in zip(V.parameters, [20.264, 6.478, 0]):
            self.assertAlmostEqual(x, y, places=3)
            
    def test_sparse_maxlag_30(self):
        V = Variogram(self.c, self.v, maxlag=30)

        for x, y in zip(V.parameters, [17.128, 6.068, 0]):
            self.assertAlmostEqual(x, y, places=3)


class TestVariogramInstatiation(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (30, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 30)

    def test_standard_settings(self):
        V = Variogram(self.c, self.v)

        for x, y in zip(V.parameters, [7.122, 13.966, 0]):
            self.assertAlmostEqual(x, y, places=3)

    def test_sparse_standard_settings(self):
        V = Variogram(self.c, self.v, maxlag=10000)

        for x, y in zip(V.parameters, [7.122, 13.966, 0]):
            self.assertAlmostEqual(x, y, places=3)

    def test_input_dimensionality(self):
        c1d = np.random.normal(0, 1, 100)
        c3d = np.random.normal(0, 1, size=(100, 3))
        v = np.random.normal(10, 4, 100)

        # test 1D coords
        V = Variogram(c1d, v)
        self.assertTrue(V.dim == 1)

        # test 3D coords
        V2 = Variogram(c3d, v)
        self.assertTrue(V2.dim == 3)

    def test_pass_median_maxlag_on_instantiation(self):
        np.random.seed(1312)
        c = np.random.gamma(5, 1, (50, 2))

        np.random.seed(1312)
        v = np.random.weibull(5, 50)

        V = Variogram(c, v, maxlag='median', n_lags=4)
        bins = [0.88, 1.77, 2.65, 3.53]

        for b, e in zip(bins, V.bins):
            self.assertAlmostEqual(b, e, places=2)

    def test_pass_mean_maxlag_on_instantiation(self):
        V = Variogram(self.c, self.v, maxlag='mean', n_lags=4)

        bins = [4.23, 8.46, 12.69, 16.91]

        for b, e in zip(bins, V.bins):
            self.assertAlmostEqual(b, e, places=2)

    def test_unknown_binning_func(self):
        with self.assertRaises(ValueError) as e:
            Variogram(self.c, self.v, bin_func='notafunc')

        self.assertEqual(
            "'notafunc' is not a valid estimator for `bins`",
            str(e.exception)
        )

    def test_invalid_binning_func(self):
        with self.assertRaises(AttributeError) as e:
            V = Variogram(self.c, self.v)
            V.set_bin_func(42)

        self.assertTrue('of type string' in str(e.exception))

    def test_unknown_model(self):
        with self.assertRaises(ValueError) as e:
            Variogram(self.c, self.v, model='unknown')

        self.assertEqual(
            'The theoretical Variogram function unknown is not understood, please provide the function',
            str(e.exception)
        )

    def test_unsupported_n_lags(self):
        with self.assertRaises(ValueError) as e:
            Variogram(self.c, self.v, n_lags=15.7)

        self.assertEqual(
            'n_lags has to be a positive integer',
            str(e.exception)
        )

    def test_value_warning(self):
        with self.assertWarns(Warning) as w:
            Variogram(self.c, [42] * 30, fit_method='lm')

        self.assertEqual(
            'All input values are the same.',
            str(w.warning)
        )

    def test_value_error_on_set_trf(self):
        """Test the Attribute error when switching to TRF on single value input"""
        with self.assertRaises(AttributeError) as e:
            v = Variogram(self.c, [42] * 30, fit_method='lm')
            v.fit_method = 'trf'

        self.assertTrue("'trf' is bounded and therefore" in str(e.exception))
    
    def test_value_error_trf(self):
        """Test the Attribute error on TRF instantiation on single value input"""
        with self.assertRaises(AttributeError) as e:
            v = Variogram(self.c, [42] * 30, fit_method='trf')

        self.assertTrue("'trf' is bounded and therefore" in str(e.exception))

    def test_pairwise_diffs(self):
        """
        Test that the cross-variogram changes do not mess with the standard
        implementation of Variogram.

        """
        # build the variogram
        V = Variogram(self.c, self.v)

        # build the actual triangular distance matrix array
        diff = pdist(np.column_stack((self.v, np.zeros(len(self.v)))), metric='euclidean')

        assert_array_almost_equal(V.pairwise_diffs, diff, decimal=2)
    
    def test_pairwise_diffs_preprocessing(self):
        """
        Remove the diffs and then request the diffs again to check preprocessing
        trigger on missing pairwise residual diffs.
        """
        V = Variogram(self.c, self.v)

        # build the diffs
        diff = pdist(np.column_stack((self.v, np.zeros(len(self.v)))), metric='euclidean')

        # remove the diffs
        V._diff = None

        # check preprocessing
        assert_array_almost_equal(V.pairwise_diffs, diff, decimal=2)


class TestVariogramArguments(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (30, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 30)

    def test_binning_method_setting(self):
        V = Variogram(self.c, self.v, n_lags=4)

        # lags
        even = [10.58, 21.15, 31.73, 42.3]
        uniform = [10.25, 16.21, 22.71, 42.3]

        # test even
        assert_array_almost_equal(even, V.bins, decimal=2)

        # set to uniform
        V.set_bin_func('uniform')
        assert_array_almost_equal(uniform, V.bins, decimal=2)

        # restore even
        V.bin_func = 'even'
        assert_array_almost_equal(even, V.bins, decimal=2)

    def test_binning_method_scott(self):
        V = Variogram(self.c, self.v, bin_func='scott')

        # scott should yield 11 bins here
        self.assertTrue(V.n_lags == 11)

        assert_array_almost_equal(
            V.bins,
            np.array([4.9, 8.6, 12.4, 16.1, 19.9, 23.6, 27.3, 31.1, 34.8, 38.6, 42.3]),
            decimal=1
        )

    def test_binning_method_stable(self):
        V = Variogram(self.c, self.v, bin_func='stable_entropy')

        assert_array_almost_equal(
            V.bins,
            np.array([4.3, 8.4, 12.8, 17.1, 21.4, 25.2, 29.9, 33.2, 38.5, 42.8]),
            decimal=1
        )

    def test_binning_method_stable_maxiter(self):
        # increase maxiter - the result should stay the same
        V = Variogram(self.c, self.v, bin_func='stable_entropy', binning_maxiter=20000)

        assert_array_almost_equal(
            V.bins,
            np.array([4.3, 8.4, 12.8, 17.1, 21.4, 25.2, 29.9, 33.2, 38.5, 42.8]),
            decimal=1
        )

    def test_binning_method_stable_fix_bins(self):
        # use 50 bins over the sqrt method - this should change the bins
        V = Variogram(
            self.c,
            self.v,
            bin_func='stable_entropy',
            binning_entropy_bins=50
        )

        assert_array_almost_equal(
            V.bins,
            np.array([4.2, 8.6, 12.8, 17.1, 21.2, 25.5, 29.3, 33.2, 37.4, 43.]),
            decimal=1
        )

    def test_binning_change_nlags(self):
        V = Variogram(self.c, self.v, n_lags=5)

        # 5 lags are awaited
        self.assertTrue(V.n_lags == 5)

        # switch to fd rule
        V.bin_func = 'fd'

        self.assertTrue(V.n_lags == 13)

    def test_set_bins_directly(self):
        V = Variogram(self.c, self.v, n_lags=5)

        # set bins by hand
        bins = np.array([4., 20., 21., 25., 40.])
        V.bins = bins

        # test setting
        assert_array_almost_equal(bins, V.bins, decimal=8)

        # test cov settings
        self.assertIsNone(V.cov)
        self.assertIsNone(V.cof)

    def test_binning_callable_arg(self):

        # define a custom function similar an existing string function
        def even_func(distances, n, maxlag):
            return np.linspace(0, np.min(np.nanmax(distances), maxlag), n + 1)[1:], None

        # run custom function and string function
        V = Variogram(self.c, self.v, n_lags=8, bin_func=even_func)
        V2 = Variogram(self.c, self.v, n_lags=8, bin_func='even')

        # check the binning is indeed the same
        assert np.array_equal(V.bins, V2.bins)

    def test_binning_iterable_arg(self):

        # define a custom iterable with bin edges
        custom_bins = np.linspace(5,50,5)

        # check that the bins are set according to those edges
        V = Variogram(self.c, self.v, bin_func=custom_bins)

        assert np.array_equal(V.bins, custom_bins)
        assert V.n_lags == len(custom_bins)
        assert V.maxlag == max(custom_bins)

        # check that custom bins have priority over nlags and maxlag
        V = Variogram(self.c, self.v, bin_func=custom_bins, nlags=1000)

        assert np.array_equal(V.bins, custom_bins)
        assert V.n_lags == len(custom_bins)
        assert V.maxlag == max(custom_bins)

        V = Variogram(self.c, self.v, bin_func=custom_bins, maxlag=1000)

        assert np.array_equal(V.bins, custom_bins)
        assert V.n_lags == len(custom_bins)
        assert V.maxlag == max(custom_bins)

    def test_binning_kmeans_method(self):
        V = Variogram(
            self.c,
            self.v,
            n_lags=6,
            bin_func='kmeans',
            binning_random_state=1306
        )

        assert_array_almost_equal(
            V.bins,
            np.array([2.5, 7.7, 12.9, 18.1, 23.7, 30.3]),
            decimal=1
        )

    def test_binning_ward_method(self):
        V = Variogram(self.c, self.v, n_lags=6, bin_func='ward')

        assert_array_almost_equal(
            V.bins,
            np.array([2.5,  7.1, 11.1, 16.2, 23., 30.]),
            decimal=1
        )


    def test_estimator_method_setting(self):
        """
        Only test if the estimator functions are correctly set. The
        estimator functions themselves are tested in a unittest of their own.
        """
        V = Variogram(self.c, self.v, n_lags=4)

        estimator_list = ('cressie', 'matheron', 'dowd', 'genton', 'minmax',
                      'percentile', 'entropy')

        for estimator in estimator_list:
            # set the estimator
            V.estimator = estimator
            imported_estimator = getattr(estimators, estimator)
            self.assertEqual(imported_estimator, V.estimator)

    def test_set_estimator_wrong_type(self):
        V = Variogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_estimator(45)
        self.assertEqual(
            str(e.exception),
            'The estimator has to be a string or callable.'
        )

    def test_set_unknown_estimator(self):
        V = Variogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_estimator('notaestimator')

        self.assertEqual(
            str(e.exception),
            'Variogram estimator notaestimator is not understood, please ' +
            'provide the function.'
        )

    def test_set_dist_func(self):
        V = Variogram([(0, 0), (4, 1), (1, 1)], [1, 2, 3], n_lags=2)

        # use Manhattan distance
        V.set_dist_function('cityblock')
        for d, v in zip([5., 2., 3.], V.distance):
            self.assertEqual(d, v)

    def test_unknown_dist_func(self):
        V = Variogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_dist_function('notadistance')
        
        self.assertEqual(
            str(e.exception),
            'Unknown Distance Metric: notadistance'
        )

    def test_wrong_dist_func_input(self):
        V = Variogram(self.c, self.v)

        with self.assertRaises(ValueError) as e:
            V.set_dist_function(55)
            
        self.assertEqual(
            str(e.exception),
            'Input not supported. Pass a string or callable.'
        )

    def test_callable_dist_function(self):
        V = Variogram([(0, 0), (4, 1), (1, 1)], [1, 2, 3], n_lags=2)

        def dfunc(u, v):
            return 1

        V.set_dist_function(dfunc)

        # test
        self.assertEqual(V.dist_function, dfunc)
        self.assertTrue((V.distance==1).all())
        self.assertEqual(V.distance_matrix.shape, (3, 3))

    @staticmethod
    def disabled_test_direct_dist_setting():
        # Distance can no longer be explicitly set
        # it would require setting the whole MetricSpace, with a
        # non-sparse diagonal matrix
        
        V = Variogram([(0, 0), (4, 1), (1, 1)], [1, 2, 3], n_lags=2)

        V.distance = np.array([0, 0, 100])

        assert_array_almost_equal(V.distance, [0, 0, 100], decimal=0)

    def test_maxlag_setting_as_max_ratio(self):
        V = Variogram(self.c, self.v)

        # set maxlag to 60% of maximum distance
        V.maxlag = 0.6
        self.assertEqual(V.maxlag, np.max(V.distance) * 0.6)
        self.assertAlmostEqual(V.maxlag, 25.38, places=2)

    def test_maxlag_custom_value(self):
        V = Variogram(self.c, self.v)

        V.maxlag = 33.3
        self.assertAlmostEqual(V.maxlag, 33.3, places=1)

    def test_use_nugget_setting(self):
        V = Variogram(self.c, self.v, normalize=True)

        # test the property and setter
        self.assertEqual(V.use_nugget, False)
        self.assertEqual(V.describe()['nugget'], 0)

        # set the nugget
        V.use_nugget = True
        self.assertEqual(V.use_nugget, True)
        self.assertEqual(V._use_nugget, True)
        self.assertAlmostEqual(
            V.describe()['normalized_nugget'],
            291.28,
            places=2
        )

    def test_use_nugget_exception(self):
        with self.assertRaises(ValueError) as e:
            Variogram(self.c, self.v, use_nugget=42)

        self.assertEqual(
            str(e.exception),
            'use_nugget has to be of type bool.'
        )

    def test_n_lags_change(self):
        V = Variogram(self.c, self.v, n_lags=10)

        self.assertEqual(len(V.bins), 10)
        V.n_lags = 5
        self.assertEqual(len(V.bins), 5)

    def test_n_lags_exception(self):
        for arg in [15.5, -5]:
            with self.assertRaises(ValueError) as e:
                Variogram(self.c, self.v, n_lags=arg)

            self.assertEqual(
                str(e.exception),
                'n_lags has to be a positive integer'
            )

    def test_n_lags_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            Variogram(self.c, self.v, n_lags='auto')

    def test_set_values(self):
        V = Variogram(self.c, self.v)

        # create a new array of same length
        _old_vals = V.values
        new_vals = np.random.normal(10, 2, size=len(_old_vals))

        V.values = new_vals

        # values.setter will call set_values
        assert_array_almost_equal(V.values, new_vals, decimal=4)

    def test_value_matrix(self):
        vals = np.array([1, 2, 3, 4])
        mat = np.asarray([[0, 1, 2, 3], [1, 0, 1, 2],[2, 1, 0, 1], [3, 2, 1, 0]], dtype=int)

        V = Variogram(self.c[:4], vals)

        assert_array_almost_equal(V.value_matrix, mat, decimal=1)

    def _test_normalize_setter(self):
        # TODO: I should fix this behavior
        V = Variogram(self.c, self.v, normalize=False)

        # make sure biggest bin larger than 1.0
        self.assertGreater(np.max(V.bins), 1.0)

        # normalize
        V.normalize = True

        # now, biggest bin should be almost or exactly 1.0
        self.assertLessEqual(np.max(V.bins), 1.0)

    def test_distance_matrix(self):
        coor = [[0, 0], [1, 0], [0, 1], [1, 1]]
        vals = [0, 1, 2, 3]
        dist_mat = np.asarray([
            [0, 1, 1, 1.414],
            [1, 0, 1.414, 1],
            [1, 1.414, 0, 1],
            [1.414, 1, 1, 0]
        ])

        V = Variogram(coor, vals)

        assert_array_almost_equal(V.distance_matrix, dist_mat, decimal=3)

    def test_entropy_as_estimator(self):
        """
        Note: This unittest will change in future, as soon as the
        bin edges for Entropy calculation can be set on instantiation

        """
        V = Variogram(self.c, self.v, estimator='entropy', n_lags=10)

        assert_array_almost_equal(
            V.experimental,
            [2.97, 3.3, 3.45, 2.95, 3.33, 3.28, 3.31, 3.44, 2.65, 1.01],
            decimal=2
        )

    def test_metric_space_property(self):
        """
        Test that the MetricSpace is correctly returned
        """
        V = Variogram(self.c, self.v)

        # get the metric space through property
        mc = V.metric_space

        # assert the coords are actually the same
        assert_array_almost_equal(
            mc.coords,
            V.coordinates,
            decimal=5
        )

    def test_metric_space_readonly(self):
        """
        Verify that metric_space is a read-only property.
        """
        V = Variogram(self.c, self.v)

        with self.assertRaises(AttributeError) as e:
            V.metric_space = self.c

            self.assertTrue('read-only' in str(e.exception))

    def test_nofit(self):
        """
        Verify that providing no fitting method skips the fitting procedure
        """
        V = Variogram(self.c, self.v, fit_method=None)

        assert V.fit_method is None
        assert V.cov is None
        assert V.cof is None

    def test_get_bin_count(self):

        V = Variogram(self.c, self.v)

        # check type
        assert isinstance(V.bin_count, np.ndarray)

        # check against real bin count
        assert np.array_equal(V.bin_count, np.array([22, 54, 87, 65, 77, 47, 46, 24, 10,  2]))

        # check property gets updated
        old_bin_count = V.bin_count

        # when setting binning function
        V.bin_func = 'uniform'
        assert not np.array_equal(V.bin_count, old_bin_count)

        # when setting maxlag
        old_bin_count = V.bin_count
        V.maxlag = 25
        assert not np.array_equal(V.bin_count, old_bin_count)

        # when setting nlags
        V.n_lags = 5
        assert len(V.bin_count) == 5


class TestVariogramFittingProcedure(unittest.TestCase):
    def setUp(self):
        np.random.seed(1337)
        self.c = np.random.gamma(10, 8, (50, 3))
        np.random.seed(1337)
        self.v = np.random.normal(10, 4, 50)

        # build a standard variogram to be used
        self.V = Variogram(
            self.c, self.v, n_lags=5, normalize=False, use_nugget=True
        )

    def test_fit_sigma_is_None(self):
        self.V.fit_sigma = None

        self.assertIsNone(self.V.fit_sigma)

    def test_fit_sigma_explicit(self):
        sigs = [.8, .5, 2., 2., 5.]
        self.V.fit_sigma = sigs

        for x, y in zip(sigs, self.V.fit_sigma):
            self.assertEqual(x, y)

    def test_fit_sigma_raises_AttributeError(self):
        self.V.fit_sigma = (0, 1, 2)

        with self.assertRaises(AttributeError) as e:
            self.V.fit_sigma
        
        self.assertTrue(
            'len(fit_sigma)' in str(e.exception)
        )

    def test_fit_sigma_raises_ValueError(self):
        self.V.fit_sigma = 'notAnFunction'

        with self.assertRaises(ValueError) as e:
            self.V.fit_sigma

        self.assertTrue(
            "fit_sigma is not understood." in str(e.exception)
        )

    def test_fit_sigma_linear(self):
        self.V.fit_sigma = 'linear'

        # test the sigmas
        sigma = self.V.fit_sigma
        for s, _s in zip(sigma, [.2, .4, .6, .8, 1.]):
            self.assertAlmostEqual(s, _s, places=8)

        # test parameters:
        self.V.fit()
        assert_array_almost_equal(
            self.V.parameters, [13., 0.3, 18.], decimal=1
        )

    def test_fit_sigma_exp(self):
        self.V.fit_sigma = 'exp'

        # test the sigmas
        sigma = self.V.fit_sigma
        for s, _s in zip(sigma, [0.0067, 0.0821, 0.1889, 0.2865, 0.3679]):
            self.assertAlmostEqual(s, _s, places=4)

        # test parameters
        assert_array_almost_equal(
            self.V.parameters, [25., 0.2, 18.5], decimal=1
        )

    def test_fit_sigma_sqrt(self):
        self.V.fit_sigma = 'sqrt'

        # test the sigmas
        assert_array_almost_equal(
            self.V.fit_sigma, [0.447, 0.632, 0.775, 0.894, 1.], decimal=3
        )

        # test the parameters
        assert_array_almost_equal(
            self.V.parameters, [19.7, 1.5,  16.4], decimal=1
        )

    def test_fit_sigma_sq(self):
        self.V.fit_sigma = 'sq'

        # test the sigmas
        assert_array_almost_equal(
            self.V.fit_sigma, [0.04, 0.16, 0.36, 0.64, 1.], decimal=2
        )

        # test the parameters
        assert_array_almost_equal(
            self.V.parameters, [5.4, 0.1,  18.5], decimal=1
        )

    def test_fit_sigma_entropy(self):
        # load data sample
        data = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
        V = Variogram(
            data[['x', 'y']].values,
            data.z.values,
            n_lags=12,
            fit_method='ml',
            fit_sigma='entropy'
        )

        assert_array_almost_equal(
            V.parameters, [65.9, 1.3, 0], decimal=1
        )

    def test_fit_sigma_on_the_fly(self):
        self.V.fit(sigma='sq')

        # test the sigmas
        assert_array_almost_equal(
            self.V.fit_sigma, [0.04, 0.16, 0.36, 0.64, 1.], decimal=2
        )

        # test the parameters
        assert_array_almost_equal(
            self.V.parameters, [5.4, 0.1,  18.5], decimal=1
        )

    def test_fit_lm(self):
        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
        V = Variogram(
            df[['x', 'y']],
            df.z.values,
            use_nugget=True,
            n_lags=8, fit_method='lm'
        )

        # test the parameters
        assert_array_almost_equal(
            V.parameters, [162.3, 0.5, 0.8], decimal=1
        )

    def test_fitted_model(self):
        self.V._fit_method = 'trf'
        self.V.fit_sigma = None
        fun = self.V.fitted_model

        result = np.array([12.48, 17.2, 17.2, 17.2])

        assert_array_almost_equal(
            result, list(map(fun, np.arange(0, 20, 5))),
            decimal=2
        )

    def test_unavailable_method(self):
        with self.assertRaises(AttributeError) as e:
            self.V.fit(method='unsupported')

        self.assertTrue(
            "fit_method has to be one of" in str(e.exception)
        )

    def test_implicit_run_fit_fitted_model(self):
        self.V.fit_sigma = None
        self.V._fit_method = 'trf'
        result = np.array([12.48, 17.2, 17.2, 17.2])

        # remove cof
        self.V.cof = None

        # test on fitted model
        fun = self.V.fitted_model

        assert_array_almost_equal(
            result, list(map(fun, np.arange(0, 20, 5))), decimal=2
        )

    def test_implicit_run_fit_transform(self):
        self.V.fit_sigma = None
        self.V._fit_method = 'trf'
        result = np.array([12.48, 17.2, 17.2, 17.2])

        # test on transform
        self.V.cof = None
        res = self.V.transform(np.arange(0, 20, 5))

        assert_array_almost_equal(result, res, decimal=2)

    def test_harmonize_model(self):
        # load data sample
        data = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
        V = Variogram(data[['x', 'y']].values, data.z.values)

        V.model = 'harmonize'
        x = np.linspace(0, np.max(V.bins), 10)

        assert_array_almost_equal(
            V.transform(x),
            [np.NaN, 0.57, 1.01, 1.12, 1.15, 1.15, 1.15, 1.15, 1.21, 1.65],
            decimal=2
        )

    def test_ml_default(self):
        # load data sample
        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
        V = Variogram(
            df[['x', 'y']],
            df.z.values,
            use_nugget=True,
            n_lags=15,
            fit_method='ml'
        )

        assert_array_almost_equal(
            V.parameters, np.array([41.18, 1.2, 0.]), decimal=2
        )

    def test_ml_sq_sigma(self):
        # load data sample
        df = pd.read_csv(os.path.dirname(__file__) + '/sample.csv')
        V = Variogram(
            df[['x', 'y']],
            df.z.values,
            use_nugget=True,
            n_lags=15,
            fit_method='ml',
            fit_sigma='sq'
        )

        assert_array_almost_equal(
            V.parameters, np.array([42.72, 1.21, 0.]), decimal=2
        )

    def test_manual_fit(self):
        V = Variogram(
            self.c,
            self.v,
            fit_method='manual',
            model='spherical',
            fit_range=10.,
            fit_sill=5.
        )

        self.assertEqual(V.parameters, [10., 5., 0.0])
    
    def test_manual_fit_change(self):
        V = Variogram(
            self.c,
            self.v,
            fit_method='trf',
            model='matern',
        )

        # switch to manual fit
        V._fit_method = 'manual'
        V.fit(range=10, sill=5, shape=3)

        self.assertEqual(V.parameters, [10., 5., 3., 0.0])

    def test_manual_raises_missing_params(self):
        with self.assertRaises(AttributeError) as e:
            Variogram(self.c, self.v, fit_method='manual')
            self.assertTrue('For manual fitting' in str(e.exception))

    def test_manual_preserve_params(self):
        V = Variogram(self.c, self.v, fit_method='trf', n_lags=8)
        params = V.parameters

        # switch fit method
        V._fit_method = 'manual'
        V.fit(sill=14)

        # expected output
        params[1] = 14.

        assert_array_almost_equal(
            V.parameters,
            params,
            decimal=1
        )
    
    def test_implicit_nugget(self):
        V = Variogram(self.c, self.v, use_nugget=False)

        # no nugget used
        self.assertTrue(V.parameters[-1] < 1e-10)

        # switch to manual fitting
        V.fit(method='manual', sill=5., nugget=2.)

        self.assertTrue(abs(V.parameters[-1] - 2.) < 1e-10)


class TestVariogramQualityMeasures(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (30, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 30)

    def test_residuals(self):
        V = Variogram(self.c, self.v)
        assert_array_almost_equal(
            V.residuals,
            np.array(
                [-3.43e-08, -1.33e-01, 2.11e+00, 4.89e+00, 1.37e+00, 1.50e+00,
                 -3.83e+00, -6.89e+00, 3.54e+00, -2.55e+00]),
            decimal=2
        )

    def test_rmse(self):
        V = Variogram(self.c, self.v)

        for model, rmse in zip(
                ['spherical', 'gaussian', 'stable'],
                [3.3705, 3.3707, 3.193]
        ):
            V.set_model(model)
            self.assertAlmostEqual(V.rmse, rmse, places=4)

    def test_mean_residual(self):
        V = Variogram(self.c, self.v)

        for model, mr in zip(
            ['spherical', 'cubic', 'stable'],
            [2.6803, 2.6803, 2.6966]
        ):
            V.set_model(model)
            self.assertAlmostEqual(V.mean_residual, mr, places=4)

    def test_nrmse(self):
        V = Variogram(self.c, self.v, n_lags=15)

        for model, nrmse in zip(
            ['spherical', 'gaussian', 'stable', 'exponential'],
            [0.3536, 0.3535, 0.3361, 0.3499]
        ):
            V.set_model(model)
            self.assertAlmostEqual(V.nrmse, nrmse, places=4)

    def test_nrmse_r(self):
        V = Variogram(self.c, self.v, estimator='cressie')

        self.assertAlmostEqual(V.nrmse_r, 0.63543, places=5)

    def test_r(self):
        V = Variogram(self.c, self.v, n_lags=12, normalize=False)

        for model, r in zip(
            ('gaussian', 'exponential', 'stable'), 
            [0.39, 0.55, 0.60]
        ):
            V.set_model(model)
            self.assertAlmostEqual(V.r, r, places=2)

    def test_NS(self):
        V = Variogram(self.c, self.v, n_lags=15, normalize=False)

        for estimator, NS in zip(
            ('matheron', 'genton', 'dowd'),
            [0.0206, 0.0206, 0.0206]
        ):
            self.assertAlmostEqual(V.NS, NS, places=4)
        
    def test_mae(self):
        V = Variogram(self.c, self.v, n_lags=15)

        self.assertAlmostEqual(V.mae, 3.91, places=2)

    def test_mse(self):
        V = Variogram(self.c, self.v, n_lags=15)

        self.assertAlmostEqual(np.sqrt(V.mse), V.rmse, places=6)

    def test_update_kwargs(self):
        V = Variogram(self.c, self.v, percentile=.3)

        self.assertAlmostEqual(
            V._kwargs.get('percentile'), 0.3, places=1
        )

        # change the parameter
        V.update_kwargs(percentile=0.7)

        self.assertAlmostEqual(
            V._kwargs.get('percentile'), 0.7, places=1
        )

    def test_kwargs_setter_in_experimental(self):
        V = Variogram(self.c, self.v, estimator='percentile')

        # store with p of 50 == median
        exp = V.experimental

        V.update_kwargs(percentile=25)

        exp2 = V.experimental

        # 25% should be very different from median
        with self.assertRaises(AssertionError):
            assert_array_almost_equal(exp, exp2, decimal=2)

    def test_residuals_deprecation(self):
        """Variogram.residuals is deprecated in favor of model_residuals"""
        with self.assertWarns(DeprecationWarning) as w:
            Variogram(self.c, self.v).residuals

        self.assertTrue('residuals is deprecated and will be removed' in str(w.warning))


class TestVariogramMethods(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (30, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 30)

        self.V = Variogram(self.c, self.v, normalize=False, n_lags=10)

    def test_get_empirical(self):
        bins = self.V.bins
        exp = self.V.experimental

        emp_x, emp_y = self.V.get_empirical()

        # test
        assert_array_almost_equal(bins, emp_x)
        assert_array_almost_equal(exp, emp_y)
    
    def test_get_empirical_center(self):
        V = Variogram(self.c, self.v)

        # overwrite bins
        V.bins = [4, 8, 9, 12, 15]
        emp_x, emp_y = V.get_empirical(bin_center=True)

        assert_array_almost_equal(emp_x, [2., 6., 8.5, 10.5, 13.5])

    def test_clone_method(self):
        # copy variogram
        copy = self.V.clone()

        # test against bins and experimental
        assert_array_almost_equal(copy.experimental, self.V.experimental)
        assert_array_almost_equal(copy.bins, self.V.bins)

    def test_data_no_force(self):
        lags, var = self.V.data(n=10, force=False)

        assert_array_almost_equal(
            lags,
            [0.,  4.7,  9.4, 14.1, 18.8, 23.5, 28.2, 32.9, 37.6, 42.3], 
            decimal=2
        )

        assert_array_almost_equal(
            var,
            [0., 11.82, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97],
            decimal=2
        )

    def disabled_test_data_with_force(self):
        # Distance can no longer be explicitly set
        # it would require setting the whole MetricSpace, with a
        # non-sparse diagonal matrix
        
        # should work if _dist is corccupted
        self.V._dist = self.V._dist * 5.
        self.V.cof = None
        lags, var = self.V.data(n=10, force=True)

        assert_array_almost_equal(
            lags,
            [0., 4.7, 9.4, 14.1, 18.8, 23.5, 28.2, 32.9, 37.6, 42.3],
            decimal=2
        )

        assert_array_almost_equal(
            var,
            [0., 11.82, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97, 13.97],
            decimal=2
        )

    def test_data_normalized(self):
        V = self.V.clone()

        V.normalize = True

        lags, var = V.data(n=5, force=True)

        assert_array_almost_equal(
            lags,
            [0., 10.58, 21.15, 31.73, 42.3],
            decimal=2
        )

        assert_array_almost_equal(
            var,
            [0., 13.97, 13.97, 13.97, 13.97],
            decimal=2
        )
    
    def test_parameter_property_matern(self):
        V = self.V.clone()
        
        # test matern
        param = [42.3, 16.2, 0.1, 0.]
        V.set_model('matern')
        assert_array_almost_equal(V.parameters, param, decimal=2)
    
    def test_parameter_property_stable(self):
        V = self.V.clone()

        # test stable
        param = [42.3, 15.79, 0.45,  0.]
        V.set_model('stable')
        assert_array_almost_equal(V.parameters, param, decimal=2)


class TestVariogramPlotlyPlots(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (150, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 150)
        self.V = Variogram(self.c, self.v)

    def test_plotly_main_plot(self):
        if PLOTLY_FOUND:
            # switch to plotly
            plotting.backend('plotly')

            self.assertTrue(
                isinstance(self.V.plot(show=False), go.Figure)
            )

            plotting.backend('matplotlib')

    def test_plotly_scattergram(self):
        if PLOTLY_FOUND:
            # switch to plotly
            plotting.backend('plotly')

            self.assertTrue(
                isinstance(self.V.scattergram(show=False), go.Figure)
            )

            plotting.backend('matplotlib')

    def test_plotly_location_trend(self):
        if PLOTLY_FOUND:
            # switch to plotly
            plotting.backend('plotly')

            self.assertTrue(
                isinstance(self.V.location_trend(show=False), go.Figure)
            )

            plotting.backend('matplotlib')

    def test_plotly_dd_plot(self):
        if PLOTLY_FOUND:
            # switch to plotly
            plotting.backend('plotly')

            self.assertTrue(
                isinstance(self.V.distance_difference_plot(show=False), go.Figure)
            )

            plotting.backend('matplotlib')
    
    def test_undefined_backend(self):
        # force the backend into an undefined state
        import skgstat
        skgstat.__backend__ = 'not-a-backend'

        for fname in ('plot', 'scattergram', 'location_trend', 'distance_difference_plot'):
            with self.assertRaises(ValueError) as e:
                self.V.plot()

                self.assertEqual(
                    str(e.exception),
                    'The plotting backend has an undefined state.'
                )
        
        # make the backend valid again
        skgstat.__backend__ = 'matplotlib'


class TestSampling(unittest.TestCase):
    def setUp(self):
        self.data = pd.read_csv(os.path.dirname(__file__) + '/pan_sample.csv')

    def test_full_vs_full_sample(self):
        Vf = Variogram(
            self.data[['x', 'y']].values,
            self.data.z.values,
            binning_random_state=44).describe()

        Vs = Variogram(
            self.data[['x', 'y']].values,
            self.data.z.values, samples=len(self.data),
            binning_random_state=44).describe()

        self.assertAlmostEqual(Vf["normalized_effective_range"], Vs["normalized_effective_range"], delta = Vf["normalized_effective_range"] / 10)
        self.assertAlmostEqual(Vf["effective_range"], Vs["effective_range"], delta = Vf["effective_range"] / 10)
        self.assertAlmostEqual(Vf["sill"], Vs["sill"], delta = Vf["sill"] / 10)

    def test_samples(self):
        Vf = Variogram(
            self.data[['x', 'y']].values,
            self.data.z.values, samples=len(self.data),
            binning_random_state=44).describe()

        for sample_size in np.linspace(0.5, 1., 10):
            Vs = Variogram(
                self.data[['x', 'y']].values,
                self.data.z.values, samples=sample_size,
                binning_random_state=44).describe()
        
            self.assertAlmostEqual(Vf["normalized_effective_range"], Vs["normalized_effective_range"], delta = Vf["normalized_effective_range"] / 5)
            self.assertAlmostEqual(Vf["effective_range"], Vs["effective_range"], delta = Vf["effective_range"] / 5)
            self.assertAlmostEqual(Vf["sill"], Vs["sill"], delta = Vf["sill"] / 5)


class TestVariogramPickling(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (150, 2))
        np.random.seed(42)
        self.v = np.random.normal(10, 4, 150)

        self.V = Variogram(self.c, self.v, normalize=False)

    def test_save_load_pickle(self):
        """
        Only test if loading and saving a pickle works without error
        """
        pickle.loads(pickle.dumps(self.V))
        return True


class TestCrossVariogram(unittest.TestCase):
    def setUp(self):
        # set up default values, whenever c and v are not important
        np.random.seed(42)
        self.c = np.random.gamma(10, 4, (100, 2))
        np.random.seed(42)
        self.v = np.random.multivariate_normal([10.5, 5.6], [[1.5, 3.2], [3.2, 1.5]], size=100)

    def test_cross_variable_dim_error(self):
        """Generate values of too high dimensionality"""
        # generate 3D values
        np.random.seed(42)
        v = np.random.normal(10, 4, (100, 3, 3))

        with self.assertRaises(ValueError) as e:
            Variogram(self.c, v)

        self.assertTrue('values has to be 1d (classic variogram)' in str(e.exception))

    def test_too_many_cross_variables(self):
        """Generate too many co-variables"""
        np.random.seed(42)
        v = np.random.normal(10, 3, (100, 3))

        with self.assertRaises(ValueError) as e:
            Variogram(self.c, v)

        self.assertTrue('create a grid of cross-variograms' in str(e.exception))

    def test_cross_instantiation(self):
        """Create a cross-variogram without error"""
        vario = Variogram(self.c, self.v, maxlag='median')

        self.assertTrue(vario.is_cross_variogram)

    def test_cross_shapes(self):
        """Check that the cross-variogram does not change the diff shape"""
        vario = Variogram(self.c, self.v, maxlag='median')

        self.assertTrue(vario.pairwise_diffs.ndim == 1)

    def test_covariable(self):
        """check that the covariable was correctly instantiated"""
        vario = Variogram(self.c, self.v, maxlag='median')

        self.assertTrue(vario.is_cross_variogram)
        assert_array_almost_equal(self.v[:,1], vario._co_variable)

    def test_directional_instantiation(self):
        """Check that the directional variogram is also instantiated."""
        vario = DirectionalVariogram(self.c, self.v, maxlag='median')

        self.assertTrue(vario.pairwise_diffs.ndim == 1)

    def test_directional_covariable(self):
        """Check that the directional variogram instantiated the covariable correctly"""
        vario = DirectionalVariogram(self.c, self.v, maxlag='median')

        self.assertTrue(vario.is_cross_variogram)
        assert_array_almost_equal(self.v[:,1], vario._co_variable)


if __name__ == '__main__':  # pragma: no cover
    os.environ['SKG_SUPRESS'] = 'TRUE'
    unittest.main()



from .Variogram import Variogram
from .DirectionalVariogram import DirectionalVariogram
from .SpaceTimeVariogram import SpaceTimeVariogram
from .Kriging import OrdinaryKriging
from .MetricSpace import MetricSpace, MetricSpacePair, ProbabalisticMetricSpace, RasterEquidistantMetricSpace
from . import interfaces
from . import data
from . import util

# set some stuff
__version__ = '1.0.5'
__author__ = 'Mirko Maelicke <mirko.maelicke@kit.edu>'
__backend__ = 'matplotlib'



import pytest
import numpy as np
import skgstat as skg
import scipy

# produce a random dataset
np.random.seed(42)
rcoords = np.random.gamma(40, 10, size=(500, 2))
np.random.seed(42)
rvals = np.random.normal(10, 4, 500)

def test_invalid_dist_func():
    # instantiate metrix space
    ms = skg.MetricSpace(rcoords, dist_metric='euclidean')

    with pytest.raises(AttributeError) as e:
        skg.Variogram(ms, rvals, dist_func='cityblock')

        assert 'Distance metric' in e.value


def test_sparse_matrix_no_warning():
    # make a really sparse matrix
    sparse = skg.MetricSpace(rcoords, max_dist=5)

    # call triangular_distance_matrix without warning
    V = skg.Variogram(sparse, rvals)
    V.triangular_distance_matrix


def test_dense_matrix_warning():
    dense = skg.MetricSpace(rcoords)

    # check the warning
    with pytest.raises(RuntimeWarning) as w:
        V = skg.Variogram(dense, rvals)
        V.triangular_distance_matrix

        assert 'Only available' in w.value


def test_unknown_metric():
    with pytest.raises(ValueError) as e:
        skg.MetricSpace(rcoords, dist_metric='foobar')

        assert 'Unknown Distance Metric:' in e.value


def test_tree_non_euklidean():
    with pytest.raises(ValueError) as e:
        ms = skg.MetricSpace(rcoords, 'cityblock')
        ms.tree

        assert 'can only be constructed' in e.value


def test_metric_pair_metrix():
    c1 = np.random.gamma(100, 4, (300, 2))
    c2 = np.random.gamma(50, 5, (100, 2))
    ms1 = skg.MetricSpace(c1, dist_metric='cityblock')
    ms2 = skg.MetricSpace(c2, dist_metric='euclidean')

    with pytest.raises(ValueError) as e:
        skg.MetricSpacePair(ms1, ms2)

        assert 'same distance metric' in e.value


def test_metric_pair_max_dist():
    c1 = np.random.gamma(100, 4, (300, 2))
    c2 = np.random.gamma(50, 5, (100, 2))
    ms1 = skg.MetricSpace(c1, max_dist=50)
    ms2 = skg.MetricSpace(c2, max_dist=400)

    with pytest.raises(ValueError) as e:
        skg.MetricSpacePair(ms1, ms2)

        assert 'same max_dist' in e.value

def test_raster_metric():

    # Generate a gridded dataset
    shape = (100, 100)
    np.random.seed(42)
    vals = np.random.normal(0, 1, size=shape)

    # Coordinates
    x = np.arange(0, shape[0])
    y = np.arange(0, shape[1])
    xx, yy = np.meshgrid(x, y)

    # Flatten everything because we don't care about the 2D at this point
    coords = np.dstack((xx.flatten(), yy.flatten())).squeeze()
    vals = vals.flatten()

    # Run the computation
    rems = skg.RasterEquidistantMetricSpace(coords, shape=shape, extent=(x[0],x[-1],y[0],y[-1]), samples=10, runs=10,
                                            rnd=42, verbose=True)

    # Minimal check of the output
    assert rems.max_dist == pytest.approx(140,rel=0.01)
    assert rems.res == pytest.approx(1, rel=0.0001)
    assert isinstance(rems.dists, scipy.sparse.csr.csr_matrix)
    assert rems.dists.shape == (10000, 10000)

    # Check the random state provides the same final center
    assert all(rems._centers[-1] == np.array([62, 52]))

    # Check the interface with a Variogram object works
    V = skg.Variogram(rems, vals)

    assert V.bin_count is not None
    # Check the variogram is always the same with the random state given
    assert V.experimental[0] == pytest.approx(0.89,0.01)

    # Check that the routines are robust to very few data points in the grid (e.g., from nodata values)
    coords_sub = coords[0::1000]
    vals_sub = vals[0::1000]
    rems_sub = skg.RasterEquidistantMetricSpace(coords_sub, shape=shape, extent=(x[0],x[-1],y[0],y[-1]), samples=100, runs=10,
                                            rnd=42)
    V = skg.Variogram(rems_sub, vals_sub)

    # Check with a single isolated point possibly being used as center
    coords_sub = np.concatenate(([coords[0]], coords[-10:]))
    vals_sub = np.concatenate(([vals[0]], vals[-10:]))
    rems_sub = skg.RasterEquidistantMetricSpace(coords_sub, shape=shape, extent=(x[0],x[-1],y[0],y[-1]), samples=100, runs=11,
                                            rnd=42)
    V = skg.Variogram(rems_sub, vals_sub)



"""
This module implements a maximum likelihood function for variogram models.
The definition is taken from [601]_:

References
----------
[601]   Lark, R. M. "Estimating variograms of soil properties by the 
        method‐of‐moments and maximum likelihood." European Journal 
        of Soil Science 51.4 (2000): 717-728.
"""
from typing import Callable, List
from itertools import cycle

import numpy as np
from scipy.spatial.distance import squareform
from scipy.linalg import inv, det

from skgstat import Variogram


DOC_TEMPLATE = """Autocorrelation function.
This function calcualtes the sptial autocorrelation for any
model function only, by setting nugget to 0 and sill to 1.
This can be used to create an autocorreation matrix as used
to derive a maximum likelihhod function for the model.

Original documentation:
{doc}
"""

LIKE_DOC = """
Negative log-likelihood function following [601].
This likelihood function is based on a ``{model_name}`` model.
The sample of the variogram instance contained ``{n}`` points.

References
----------
[601]   Lark, R. M. "Estimating variograms of soil properties by the 
        method‐of‐moments and maximum likelihood." European Journal 
        of Soil Science 51.4 (2000): 717-728.
"""


def _model_transformed(model_func, has_s: bool = False) -> Callable:
    """
    Transforms the model parameter input to fit likelihood function
    input parameters.
    The returned function can be used to create the log-likelihood
    function that has to be minimized.
    To build up the spatial autocorrelation matrix, the spatial
    autocorrelation has to be separated from nugget and sill.
    """
    if has_s:
        def wrapped(h, r, s):
            return model_func(h, r, s=s, c0=1, b=0)
    else:
        def wrapped(h, r):
            return model_func(h, r, c0=1, b=0)

    # add the original docstring
    wrapped.__doc__ = DOC_TEMPLATE.format(doc=model_func.__doc__)
    wrapped.__name__ = f"autocorr_{model_func.__name__}"

    return wrapped


def _build_A(transformed_func: Callable, params: List[float], dists: np.ndarray) -> np.ndarray:
    """
    Builds the autocorrelation matrix for a given model function.
    """
    if len(params) == 4:
        r, c0, s, b = params
        a = np.fromiter(map(transformed_func, dists, cycle([r]), cycle([s])), dtype=float)
    else:
        r, c0, b = params
        # calcualte the upper triangle of A:
        a = np.fromiter(map(transformed_func, dists, cycle([r])), dtype=float)

    # build the full matrix
    A = squareform((c0 / (c0 + b)) * (1 - a))

    # replace diagonal 0 with ones
    np.fill_diagonal(A, 1)

    return A


def get_likelihood(variogram: Variogram) -> Callable:
    """
    Returns the maximum likelihood function for the given variogram.
    The function takes a list of model parameters and returns the
    negative log-likelihood for this set of parameters.
    The signature of the reutrned function has the interface as
    needed by :func:`scipy.optimize.minimize`.
    The paramters for the models are given as ``[r, c0, b]``, which
    are the effective range, sill and nugget. For the
    :func:`Matérn <skgstat.models.matern>` and
    :func:`Stable <skgstat.models.stable>` models, the parameters include
    an additional parameter ``s``, which is the smoothness parameter.

    Example
    -------

    .. code-block:: python

        from scipy.optimize import minimize
        like = get_likelihood(variogram)
        res = minimize(like)

    It is, hwowever, highly recommended to provide an initial guess and
    specify bounds for the parameter space.

    """
    # extract the current data
    values = variogram.values
    # TODO: there is a bug as this is not working:
    # dists = variogram.distance
    try:
        dists = squareform(variogram._X.dists.todense())
    except AttributeError:
        dists = variogram.distance

    # get the transformed model func
    has_s = variogram.model.__name__ in ('matern', 'stable')
    transformed_func = _model_transformed(variogram.model, has_s=has_s)

    def likelihood(params: List[float]) -> float:
        # calculate A
        A = _build_A(transformed_func, params, dists)
        n = len(A)

        # invert A
        A_inv = inv(A)

        # build the 1 vector and t
        ones = np.ones((n, 1))
        z = values.reshape(n, -1)

        # build the estimate matrix for field means
        m = inv(ones.T @ A_inv @ ones) @ (ones.T @ A_inv @ z)
        b = np.log((z - m).T @ A_inv @ (z - m))

        # get the log of determinant of A
        D = np.log(det(A))
        # np.log(0.0) is -inf, so we need to check for this
        if D == -np.inf:  # pragma: no cover
            # print a warning and return np.inf to not use these parameters
            print("Warning: D = -inf, returning np.inf")
            return np.inf

        # finally log-likelihood of the model given parameters
        loglike = (n / 2)*np.log(2*np.pi) + (n / 2) - (n / 2) * np.log(n) + 0.5 * D + (n / 2) * b

        # this is actually a 1x1 matrix
        return loglike.flatten()[0]

    # add the docstring to the likelihood function
    likelihood.__doc__ = LIKE_DOC.format(model_name=variogram.model.__name__, n=len(values))

    # return the likelikhood function
    return likelihood
